
import collections, os, sys, time
import pickle, traceback, shutil
import readcfg, ilp, ipet, test, pickleserver
from twisted.internet import reactor, defer, protocol
from readcfg import CFG_Vertex, CFG_Edge
from config import M5_BIN, GCC_BASE_DIR, SP

MAIN_MEMORY_ACCESS_COST = 50
[OPEN, DURING, CLOSE] = [1, 2, 4]
[OC_TYPE, USER_TYPE] = "OU"
TOLERANCE = 20
MAX_FAILURES = 20
            
class My_CFG_Edge(readcfg.CFG_Edge):
    def __init__(self, *args):
        readcfg.CFG_Edge.__init__(self, *args)
        self.loop_number = None
        self.live_set = 0
        self.connection_region = None
        self.alloc_obj = None
        self.open_close_cost = 0


class My_Base_Creator(readcfg.CFG_Base_Creator):
    def __init__(self, *args):
        readcfg.CFG_Base_Creator.__init__(self, *args)
        self.live_vertexes_list = []
        self.live_vertexes_dict = dict()
        self.live_edges_list = []
        self.live_edges_dict = dict()

    def Size(self):
        return self.size * self.max

    def Tx_Cost(self):
        return (self.size / 4) + MAIN_MEMORY_ACCESS_COST

    def Live_During(self, edge):
        return (self.live_edges_dict.get(edge.id, 'D') != 'D')

    def Important(self):
        return self.type == readcfg.BC_Heap

    def Info(self):
        return '%x max %d size %d creators %d users %d live %d' % (
                self.creators[ 0 ].pc, self.max, 
                self.size, len(self.creators), 
                len(self.users), len(self.live_vertexes_list))

    def Minimal_Liveness_Region(self):
        self.live_vertexes_list = []
        self.live_vertexes_dict = dict()
        self.live_edges_list = []
        self.live_edges_dict = dict()
        edge_status = dict()

        def DFS(edge):
            s = edge_status.get(edge.id, 'X')
            if (s == 'L'):
                return True # live
            elif (s in 'D'):
                return False # dead
               
            edge_status[ edge.id ] = '?'
            live = False

            if (s != '?'):
                # not in progress.. is it used below this point?
                for e2 in edge.end.exits:
                    live = DFS(e2) or live

            uses = False
            for (bc, pc_list) in edge.user:
                if (self == bc):
                    uses = True
                    break

            if (uses):
                # must be live, used here
                live = True
            elif (self in edge.creator):
                # must be dead, creates but does not use
                live = False

            if (s != '?'):
                # not in progress.. mark final state of edge
                if (live):
                    edge_status[ edge.id ] = 'L'
                    edge.start.present.append(self)
                    self.live_vertexes_list.append(edge.start)
                    self.live_vertexes_dict[ edge.start.id ] = 'L'
                    self.live_edges_list.append(edge)
                    self.live_edges_dict[ edge.id ] = 'L'
                else:
                    edge_status[ edge.id ] = 'D'

            return live

        for creator in self.creators:
            DFS(creator)

def Divide_By_Loops(cfg_entrance, debug_dir, vertex_list,
            edge_list, loop_header_list, Log):

    for edge in edge_list:
        edge.loop_number = None

    # Sort loops by distance from root.
    # Loop A can only be inside loop B if A.depth > B.depth
    loop_header_list = loop_header_list[:]
    loop_header_list.sort(key=lambda x: x.end.depth)
    number_of_loops = 0
    headers_seen = set()

    # Mark loop_number on each known element of each loop.
    for (i, header) in enumerate(loop_header_list):
        assert not header.id in headers_seen
        headers_seen.add(header.id)
        assert number_of_loops == i
        for edge in header.loop_body:
            edge.loop_number = number_of_loops
        number_of_loops += 1

    # Find expansion points - what loop are these part of?
    expansion_point = []
    for (i, header) in enumerate(loop_header_list):
        for edge1 in header.loop_body:
            assert edge1.loop_number != None
            assert edge1.loop_number >= i
            for edge2 in edge1.end.exits:
                if (edge2.loop_number == None):
                    expansion_point.append((edge1.loop_number, header,
                            edge2))

    # The entrance is also an expansion point
    for edge in cfg_entrance.exits:
        expansion_point.append((number_of_loops, None, edge))

    # Do the initial expansion, assuming that all branches leaving
    # a loop are part of that loop (until another loop is reached).
    # This is necessary because we don't yet know which loops contain
    # which other loops.
    for (loop_number, header, edge) in expansion_point:
        Expand_From(loop_number, edge, [], header, loop_number)

    # Compute relationship between loops
    loop_contained_by = dict()
    loop_contains = dict()
    for (i, header) in enumerate(loop_header_list):
        numbers_present = set()
        for edge in header.loop_body:
            assert edge.loop_number < number_of_loops
            numbers_present.add(edge.loop_number)

        assert i in numbers_present, (
            "loop %d not present at all? %d in body" % (i, 
                len(header.loop_body)))

        numbers_present = list(numbers_present)
        numbers_present.sort()
        container = numbers_present[ 0 ]
        assert container == i

        for within in numbers_present[ 1: ]:
            loop_contained_by[ within ] = max(
                container, loop_contained_by.get(within, -1))

        loop_contains[ container ] = numbers_present[ 1: ]

    # We can create a tree of loops
    Dump_Loop_Tree("%s/tree_of_loops.udg" % debug_dir, 
                loop_header_list, loop_contains)
    Dump_UDG("%s/partitioned.udg" % debug_dir, 
                vertex_list, cfg_entrance)

    # Now every edge has a number, right?
    no_number_count = 0
    for edge in edge_list:
        if (( edge.loop_number == None )
        or not (0 <= edge.loop_number <= number_of_loops)):
            no_number_count += 1

    Log('there are %d edges' % len(edge_list))
    Log("there are %d loops" % number_of_loops)
    Log('%d edges do not have numbers' % no_number_count)
    assert no_number_count == 0

    # A list of partitions can be created. Each loop has its own partition.
    partition_list = [ [] for i in xrange(number_of_loops + 1) ]
    for edge in edge_list:
        partition_list[ edge.loop_number ].append(edge)

    return partition_list


def Expand_From(loop_number, edge, contains, header, new_region):
    if (edge.loop_number != None):
        if ((edge.loop_number == loop_number)
        or (edge.loop_number in contains)):
            # within the same loop
            return loop_number
        else:
            # exit from loop, reaching a new loop
            return new_region
   
    result = new_region
    for edge2 in edge.end.exits:
        result = min(result,
            Expand_From(loop_number, edge2, contains, header, new_region))

    if (header != None):
        header.loop_body.append(edge)

    edge.loop_number = result
    return result


def Dump_UDG(name, vertex_list, cfg_entrance):
    for v in vertex_list:
        v.Clear_UDG_Flag()

    def Edge_Label(edge):
        out = []
        min_pc = 1 << 63
        max_pc = 0
        for pc in edge.path:
            min_pc = min(pc, min_pc)
            max_pc = max(pc, max_pc)

        out.append(r"id %x\n%x\n%x\n" % (edge.id, min_pc, max_pc))
        if (edge.loop_number == None):
            out.append("NO NUMBER")
        else:
            out.append("%u " % edge.loop_number)
            if (edge.loop_number < 0):
                out.append("NEGATIVE")
            else:
                out.append("loop")
        if (edge.connection_region != None):
            out.append(r"\n%u region"%  edge.connection_region)
        
        return ''.join(out)

    udg_out = file(name, "wt")
    udg_out.write('[')
    cfg_entrance.Export_UDG(udg_out, Edge_Label)
    udg_out.write(']\n')
    udg_out.close()

def Dump_Loop_Tree(name, loop_header_list, loop_contains):

    udg_out = file(name, "wt")
    udg_out.write('[')
    printed = set()

    def Print_Header(number):
        if (number in printed):
            udg_out.write(r'r("L%x")' % number)
            return

        printed.add(number)
        header = loop_header_list[ number ]
        min_pc = 1 << 63
        max_pc = 0
        for edge in header.loop_body:
            for pc in edge.path:
                min_pc = min(pc, min_pc)
                max_pc = max(pc, max_pc)

        assert max_pc > 0
        udg_out.write('l("L%x",n("C%x",[a("OBJECT",'
                    r'"%u loop\n%x\n%x")],[' % (number, number,
                    number, min_pc, max_pc))
        child = set(loop_contains[ number ])
        for child_number in loop_contains[ number ]:
            for grandchild_number in loop_contains[ child_number ]:
                child.discard(grandchild_number)
       
        child = list(child)
        child.sort()
        first = True

        for child_number in child:
            if not first:
                udg_out.write(',')
            udg_out.write('e("X%x",[a("OBJECT","")],' % (child_number))
            Print_Header(child_number)
            udg_out.write(')')
            first = False

        udg_out.write(']))')
   
    first = True
    for number in xrange(len(loop_header_list)):
        if (number in printed):
            continue
        if not first:
            udg_out.write(',')
        Print_Header(number)
        first = False

    udg_out.write(']\n')
    udg_out.close()
    assert len(printed) == len(loop_header_list)



def Make_Debug_Dir(name):
    max_length = 1
    use_temp_dir = False
    return test.Make_Debug_Dir(name, max_length, use_temp_dir)

def Optimiser(name, total_space, backtrack, limit):
    max_length = 1
    use_temp_dir = False
    debug_dir = Make_Debug_Dir(name)
    
    identifier = "%s-%u-%s-%u" % (name, total_space, str(backtrack), limit)

    optlog_file = "%s/optlog-%s" % (debug_dir, identifier)
    cfg_file = "%s/cfg2" % (debug_dir)
    oc_out_file = "%s/oc-%s" % (debug_dir, identifier)

    if ( not ((os.path.exists(cfg_file))
    and (len(file(cfg_file).read(128)) != 0))):
        return '%s - cfg2 file not found, run zirconium.py' % name

    if (os.path.exists(oc_out_file)):
        return '%s - oc out file already exists for %d' % (name, total_space)

    _log = file(optlog_file, "wt")
    _ts_base = time.time()

    def Log(*args):
        ts = "%4.0f" % (time.time() - _ts_base)
        _log.write(ts)
        print ts,
        for a in args:
            _log.write(" ")
            _log.write(str(a))
            print a,
        _log.write("\n")
        _log.flush()
        print ''

    Log(identifier)
    Log("reading CFG %s" % cfg_file)

    # Read input data & create model
    cfg_entrance = readcfg.Read_CFG(cfg_file, 
            edge_class=My_CFG_Edge,
            base_creator_class=My_Base_Creator)
    Log("identifying loops")
    loop_header_list = readcfg.Identify_Loops(cfg_entrance) # Required for IPET
    model = ilp.ILP_Solver("%s/mdl.%s.%s" % (debug_dir, name, identifier))
    (vertex_list, edge_list) = readcfg.Find_V_E(cfg_entrance)
    Log("initial solve starting")

    # Initial IPET solve (for testing, really)
    try:
        best_wcet = ipet.Simple_IPET(cfg_entrance, model)
    except ilp.ILPError, e:
        Log("Initial IPET solve failed with %s" % str(e))
        return 'initial solve failed'

    Log('all scratchpad iio wcet =', best_wcet)
    checking = best_wcet

    for edge in edge_list:
        assert edge.cost == edge.base_cost
        checking -= edge.cost * edge.frequency
    assert checking == 0

    # Discover all important base creators and put them in bc_heap_list
    bc_list = readcfg.Find_Base_Creators(cfg_entrance)
    bc_heap_list = []
    bc_lookup = dict()
    for bc in bc_list:
        if (bc.Important()):
            bc_heap_list.append(bc)
            bc_lookup[ bc.id ] = bc
    Log('found %d base creators of which %d are important' % (
            len(bc_list), len(bc_heap_list)))

    # Minimise liveness regions
    for bc in bc_heap_list:
        bc.Minimal_Liveness_Region()
        Log(bc.Info())

    Log('liveness regions are minimised')

    # Assign each edge an identifier showing what is live here
    for edge in edge_list:
        r = 0
        for (i, bc) in enumerate(bc_heap_list):
            if (bc.Live_During(edge)):
                r = r | (1 << i)
        edge.live_set = r

    Log('liveness sets created')

    # Partition graph using loop headers
    loop_partition_list = Divide_By_Loops(cfg_entrance, 
                    debug_dir, vertex_list,
                    edge_list, loop_header_list, Log)

    Log('loop partitioned into %d loop regions' % len(loop_partition_list))

    # Further partition the graph using liveness regions
    number_of_regions = 0
    for edge in edge_list:
        if (Find_All_Connected(edge, number_of_regions)):
            number_of_regions += 1

    Log("loop partitioned again into %d allocation regions" % number_of_regions)

    Dump_UDG("%s/partitioned2.udg" % debug_dir, 
                vertex_list, cfg_entrance)
        
    # One allocation object for each region
    allocation_obj_list = [ Allocation_Object(i, total_space, limit) 
            for i in xrange(number_of_regions) ]
    for edge in edge_list:
        assert edge.connection_region != None
        assert 0 <= edge.connection_region < number_of_regions
        edge.alloc_obj = allocation_obj_list[ edge.connection_region ]
        edge.alloc_obj.edge_list.append(edge)

    Log('entering core algorithm')

    # Recalculating costs
    Recalc_Costs(bc_heap_list, edge_list)
    try:
        worst_wcet = ipet.Simple_IPET(cfg_entrance, model)
    except ilp.ILPError, e:
        Log("Main memory IPET solve failed with %s" % str(e))
        return 'main solve failed'

    Log('main memory only iio wcet =', worst_wcet)
    assert worst_wcet != None
    assert worst_wcet >= best_wcet
    best_so_far_wcet = worst_wcet

    done = set()
    tolerate_bad = TOLERANCE
    failed_moves = 0
    become_intolerant_at = _ts_base + 1800
    exit_at = become_intolerant_at + 300

    undo = []

    while True:
        # What's the greatest cost that we can do something about?
        cost_list = Recalc_Costs(bc_heap_list, edge_list)
        key = None
        while len(cost_list) != 0:
            record = cost_list.pop()
            (cost, bc_id, connection_region, cost_type) = record
            key = (bc_id, connection_region, cost_type)
            if (not (key in done)):
                break

        if (key in done):
            break

        done.add(key)
        if (backtrack and (len(undo) != 0)):
            undo.append((done.discard, (key, )))
        alloc_obj = allocation_obj_list[ connection_region ]
        bc = bc_lookup[ bc_id ]
        Log("Greatest cost is %d for bc %x in region %d, cost type %s" %
                (cost, bc_id, connection_region, cost_type))

        if (cost_type == USER_TYPE):
            # Solve by changing the open state throughout this region
            assert not alloc_obj.BC_Is_Open(bc)
            if (alloc_obj.Allocate_At(bc, None)):
                Log(" Solved by opening the object in region %d" %
                        connection_region)
                undo.append((alloc_obj.Deallocate, (bc, )))
            else:
                Log(" Can't allocate this object in region %d, size is %d" %
                        (connection_region, bc.Size()))

        elif (cost_type == OC_TYPE):
            # Removing this open/close pair would be backtracking...
            # so instead try to expand it to more regions
            neighbour = set()
            for edge in alloc_obj.edge_list:
                for v in (edge.start, edge.end):
                    for elist in (v.entrances, v.exits):
                        for edge2 in elist:
                            neighbour.add(edge2.connection_region)
            neighbour.discard(connection_region)
            Log(" Neighbours are %s" % neighbour)
            for x in neighbour:
                n_alloc_obj = allocation_obj_list[ x ]
                if (n_alloc_obj.BC_Is_Open(bc)):
                    pass
                elif (n_alloc_obj.Allocate_At(bc, 
                            alloc_obj.BC_Address(bc))):
                    Log(" Now it's allocated in %d too" % x)
                    undo.append((n_alloc_obj.Deallocate, (bc, )))
                        
                else:
                    Log(" Can't spread to %d" % x)

        else:
            assert False, "Unknown cost type"

        # Recalculate costs
        Recalc_Costs(bc_heap_list, edge_list)
        try:
            new_wcet = ipet.Simple_IPET(cfg_entrance, model)
        except ilp.ILPError, e:
            Log("Intermediate optimisation IPET solve failed")
            new_wcet = worst_wcet

        improv1 = 100.0 * ((best_so_far_wcet / float(new_wcet)) - 1.0)
        improv2 = 100.0 * ((worst_wcet / float(new_wcet)) - 1.0)
        Log(' new wcet = %d ... %1.2f%% / %1.2f%%' % (new_wcet, 
                improv1, improv2))
        assert new_wcet != None
        assert new_wcet >= best_wcet

        if (new_wcet > best_so_far_wcet):
            # This is worse than the previous state
            if ((tolerate_bad > 0)
            and backtrack
            and (time.time() < become_intolerant_at)):
                # But I'll tolerate it in the hope that I get
                # something better soon
                tolerate_bad -= 1
                Log(" Tolerating bad move (%d left)" % tolerate_bad)
            else:
                for (fn, params) in undo:
                    fn(*params)

                tolerate_bad = TOLERANCE
                undo = []
                failed_moves += 1

                Log(" Undoing bad moves (%d failed moves)" % failed_moves)

                Recalc_Costs(bc_heap_list, edge_list)
                try:
                    new_wcet = ipet.Simple_IPET(cfg_entrance, model)
                except ilp.ILPError, e:
                    Log("IPET solve after undo failed")
                    return 'solve after undo failed'

                if (((failed_moves > MAX_FAILURES) and backtrack)
                or (time.time() > exit_at)):
                    Log("That's it, I think")
                    break
        else:
            # This is better, or the same
            Log(" It's a good move")
            best_so_far_wcet = new_wcet
            tolerate_bad = TOLERANCE
            undo = []

    Log("Finalising")
    for (fn, params) in undo:
        fn(*params)

    cost_list = Recalc_Costs(bc_heap_list, edge_list)
    try:
        new_wcet = ipet.Simple_IPET(cfg_entrance, model)
        assert new_wcet == best_so_far_wcet
    except ilp.ILPError, e:
        Log("Finalising IPET solve failed")
        return 'final solve failed'

    output = file(oc_out_file, "wt")
    for alloc_obj in allocation_obj_list:
        output.write("ao %x\n" % alloc_obj.id)
        for edge in alloc_obj.edge_list:
            output.write("  edge %06x\n" % edge.id)
        address = 0
        total = 0
        for (size, occupant) in alloc_obj.allocation_list:
            output.write("  alloc %04x %04x " % (address, address + size))
            if (occupant == None):
                output.write("empty\n")
            else:
                output.write("%06x\n" % occupant.id)
                total += size
            address += size
        output.write("  s %04x\n\n" % total)
    output.write("result %x %x %x\n" % (worst_wcet, new_wcet, best_wcet))
    output.close()


    Log('main memory only iio wcet =', worst_wcet)
    Log('final wcet =', new_wcet)
    Log('all scratchpad iio wcet =', best_wcet)
    improv2 = 100.0 * ((worst_wcet / float(new_wcet)) - 1.0)
    Log('SCORE %1.2f%%' % (improv2))
    return '%1.2f%%' % improv2


def Find_All_Connected(edge, region_number):
    if (edge.connection_region != None):
        # already connected
        return 0

    edge.connection_region = region_number
    total = 1
    for v in (edge.start, edge.end):
        for elist in (v.entrances, v.exits):
            for edge2 in elist:
                if ((edge2.live_set == edge.live_set)
                and (edge2.loop_number == edge.loop_number)):
                    # Connected
                    total += Find_All_Connected(edge2, region_number)

    return total

class Allocation_Object:
    def __init__(self, id, total_space, limit):
        self.id = id
        self.allocation_list = [ (total_space, None) ]
        self.limit = limit
        self.allocation_map = dict()
        self.edge_list = []

    def BC_Is_Open(self, bc):
        return self.allocation_map.has_key(bc.id)

    def BC_Address(self, bc):
        return self.allocation_map.get(bc.id, None)

    def Find_Space(self, size):
        address = 0
        for (space, occupant) in self.allocation_list:
            if ((occupant == None)
            and (space >= size)):
                return address
            address += space
        return None

    def Defragment(self):
        i = 0
        while (i < (len(self.allocation_list) - 1)):
            (space0, occupant0) = self.allocation_list[ i + 0 ]
            (space1, occupant1) = self.allocation_list[ i + 1 ]
            if (occupant0 == occupant1):
                self.allocation_list.pop(i)
                self.allocation_list[ i ] = (space0 + space1, occupant0)
            else:
                i += 1
            
        
    def Deallocate(self, bc):
        # Try to allocate this base creator 
        assert self.BC_Is_Open(bc)

        success = False
        for i in xrange(len(self.allocation_list)):
            (space, occupant) = self.allocation_list[ i ]
            if (occupant == bc):
                self.allocation_list[ i ] = (space, None) 
                del self.allocation_map[ bc.id ]
                success = True
                break
       
        assert success
        self.Defragment()
        
    def Allocate_At(self, bc, target_address=None):
        # Try to allocate this base creator 
        assert not self.BC_Is_Open(bc)

        if (len(self.allocation_map) >= self.limit):
            return False # Limit on OPEN objects is reached

        target_size = bc.Size()
        if (target_address == None):
            target_address = self.Find_Space(target_size)
        if (target_address == None):
            return False # no can do, no space
     
        low_address = 0
        for i in xrange(len(self.allocation_list)):
            (space, occupant) = self.allocation_list[ i ]
            high_address = low_address + space
            if ((target_address >= low_address)
            and (occupant == None)
            and ((target_address + target_size) <= high_address)):
                # Fits here
                space_below = target_address - low_address
                space_above = high_address - (target_address + target_size)
                assert space_below >= 0
                assert space_above >= 0

                if (space_above > 0):
                    self.allocation_list.insert(i + 1, (space_above, None))

                self.allocation_list[ i ] = (target_size, bc)

                if (space_below > 0):
                    self.allocation_list.insert(i, (space_below, None))

                self.allocation_map[ bc.id ] = target_address
                self.Defragment()
                return True
            else:
                # Does not fit here
                low_address = high_address

        return False



def Recalc_Costs(bc_list, edge_list):
    cost_map = dict()
    counted = set()

    for edge in edge_list:
        edge.cost = edge.base_cost
        edge.open_close_cost = 0

    for edge in edge_list:
        # Does this edge use any base pointers that aren't open?
        for (bc, pc_list) in edge.user:
            if (bc.Important() and (not edge.alloc_obj.BC_Is_Open(bc))):
                for pc in pc_list:
                    if not (pc in counted):
                        edge.cost += MAIN_MEMORY_ACCESS_COST
                        key = (bc.id, edge.connection_region, USER_TYPE)
                        cost_map[ key ] = (cost_map.get(key, 0) + (
                                (MAIN_MEMORY_ACCESS_COST - 1) * edge.frequency))
                        counted.add(pc)

        border = set()

        # Is this edge on the border of a region?
        for prev_edge in edge.start.entrances:
            if (prev_edge.alloc_obj != edge.alloc_obj):
                # Yes, border
                for bc in bc_list:
                    assert bc.Important()
                    before = prev_edge.alloc_obj.BC_Address(bc) 
                    after = edge.alloc_obj.BC_Address(bc) 

                    if (before != after):
                        if (before != None):
                            # Close at the end of prev_edge
                            prev_edge.cost += bc.Tx_Cost()
                            prev_edge.open_close_cost += bc.Tx_Cost()
                            key = (bc.id, prev_edge.connection_region, OC_TYPE)
                            cost_map[ key ] = (cost_map.get(key, 0) + (
                                    bc.Tx_Cost() * prev_edge.frequency))
                            border.add(bc.id)

                        if (after != None):
                            # Open at the beginning of edge
                            edge.cost += bc.Tx_Cost()
                            edge.open_close_cost += bc.Tx_Cost()
                            key = (bc.id, edge.connection_region, OC_TYPE)
                            cost_map[ key ] = (cost_map.get(key, 0) + (
                                    bc.Tx_Cost() * edge.frequency))
                            border.add(bc.id)

        # Is this edge on the border of the program?
        if len(edge.start.entrances) == 0:
            for bc in bc_list:
                assert bc.Important()
                if (edge.alloc_obj.BC_Is_Open(bc) ):
                    edge.cost += bc.Tx_Cost()
                    edge.open_close_cost += bc.Tx_Cost()
                    border.add(bc.id)
        if len(edge.end.exits) == 0:
            for bc in bc_list:
                assert bc.Important()
                if (edge.alloc_obj.BC_Is_Open(bc) ):
                    edge.cost += bc.Tx_Cost()
                    edge.open_close_cost += bc.Tx_Cost()
                    border.add(bc.id)
        
        for bc in edge.creator:
            # Non-border creators force an OPEN/CLOSE pair
            if ((not (bc.id in border))
            and (edge.alloc_obj.BC_Is_Open(bc))):
                edge.cost += 2 * bc.Tx_Cost()
                edge.open_close_cost += 2 * bc.Tx_Cost()
                

    for edge in edge_list:
        assert edge.cost >= edge.base_cost

    cost_list = []
    for ((bc_id, connection_region, cost_type), cost) in cost_map.items():
        cost_list.append((cost, bc_id, connection_region, cost_type))

    cost_list.sort()
    return cost_list

def Builder(name, subtasks):
    debug_dir = Make_Debug_Dir(name)
    cfg_file = "%s/cfg2" % (debug_dir)

    if (os.path.exists(cfg_file)
    and ( len(file(cfg_file).read(10)) == 10 )):
        return 'already done - %s' % name

    for (i, subtask) in enumerate(subtasks):
        pickle.dump(subtask, 
                file("%s/todo%d.dat" % (debug_dir, i + 1), "wb"))

    # C++ processor - identify basic blocks, WCET constraints and
    # also 
    cmd = ("./zirconium %s %s "
            "%s/bin/%s-nm %s" % (
                subtasks[ 0 ][ 'binary' ],
                debug_dir, 
                GCC_BASE_DIR, SP, os.path.dirname(M5_BIN)))
    print cmd
    rc = os.system(cmd)
    assert rc == 0
    print 'Zirconium complete'

    return 'ok'


#from specrun import Prepare_Task_Set
CANT_BUILD = "SPEC source code is required to rebuild this"
def Prepare_Task_Set():
    collect1 = pickle.load(file("s1tasks.dat", "rb"))
    task_set = dict()
    for data_in in collect1:
        name = data_in[ 0 ]
        task_set[ name ] = CANT_BUILD
    return task_set

def Prepare_Main():
    task_set = Prepare_Task_Set()
    collect1 = dict()
    collect2 = dict()
    for size in [4096, 8192, 16384]:
        for name in task_set:
            for limit in [8, 16]:
                for backtrack in [True]:
                    data_in = (name, size, backtrack, limit)
                    collect2[ data_in ] = pickleserver.TODO

    for name in task_set:
        data_in = (name, 'B', False, 0)
        collect1[ data_in ] = pickleserver.TODO
        debug_dir = Make_Debug_Dir(name)
        pickle.dump(task_set[ name ], file("%s/subtasks" % (debug_dir), "wb"))


    #pickle.dump(collect1, file("s1tasks.dat", "wb"))
    pickle.dump(collect2, file("s2tasks.dat", "wb"))

def Client(data_in):
    (name, size, backtrack, limit) = data_in
    if (size == 'B'):
        # stage 1
        debug_dir = Make_Debug_Dir(name)
        subtasks = pickle.load(file("%s/subtasks" % (debug_dir), "rb"))
        assert subtasks != CANT_BUILD, CANT_BUILD
        return Builder(name, subtasks)
    else:
        # stage 2
        assert type(size) == int
        assert size > 0
        return Optimiser(name, size, backtrack, limit)

def Client_Main():
    sys.setrecursionlimit(5000)

    server = "localhost"
    if len(sys.argv) == 2:
        server = sys.argv[ 1 ]
    elif len(sys.argv) != 1:
        print 'Error - give the server hostname as a parameter.'
        return

    def run():
        pickleserver.clientRun(server, Client)
 
    reactor.addSystemEventTrigger('after', 'startup', run)
    reactor.run()

def Main():
    size = 8192
    name = sys.argv[ 1 ]
    task_set = Prepare_Task_Set()
    print Builder(name, task_set[ name ])
    print Optimiser(name, size, True, 16)
    return 0

if ( __name__ == "__main__" ):
    sys.exit(Main())


