
# is the END operation executed or not?
# Need to be 100% clear on this point since at the moment it
# seems to depend on the context, which is clearly wrong.
# "No, it is NOT executed." -
#   Doesn't work because fetch_impl needs to be able to mark
#   the final operation with an End flag. 
# "Yes, it IS executed." -
#   That means

import sys, os, pickle, traceback, shutil, time
from twisted.internet import reactor, defer, protocol
import ilp, readcfg, ipet, pickleserver
from superblock import Superblock, Check_Superblocks, Superblock_IPET
from readcfg import BT_None, BT_Serialising, BT_Conditional, BT_Unconditional

BENCHMARKS = "benchmarks"
M5_CMD = "m5.debug"
STOP_HEURISTIC = True
NON_IDEAL_TOLERANCE = 10
MINIMUM_CHANGE_FACTOR = 0.98
STOP_HEURISTIC_POINT = 3
MAX_CANDIDATES = 3
DEFAULT_DEBUG = 2
BLACKLIST = set([ "loop3", "minmax", "fibcall" ])
[ DO_NOT_TOUCH, NOT_TRIED, STAGE_ONE ] = "DNS"

# debug_level 0 -> minimal debugging and disk space usage
# debug_level 1 -> helper log produced
# debug_level 2 -> copies of all outputs stored
# debug_level 3 -> M5 flags set for moderately verbose operation
# debug_level 4 -> M5 flags set for fully verbose operation

class M5_Bug(Exception):
    pass

def Main():
    if len(sys.argv) != 3:
        print 'Usage: %s <benchmark name> <max length>' % sys.argv[ 0 ]
        return 1


    name = sys.argv[ 1 ]
    name = name.strip()
    max_length = int(sys.argv[ 2 ])
    debug_dir = Make_Debug_Dir(name, max_length, False)
    rfd = file("%s/report.txt" % debug_dir, 'wt')

    def Report(*args):
        args = [ str(x) for x in args ]
        out = ' '.join(args) 
        print out
        rfd.write(out + "\n")
        rfd.flush()

    x = Do_Program(name, Report, DEFAULT_DEBUG, max_length, False)
    rfd.close()
    if type(x) == float:
        return 0 # ok
    else:
        return 1

def Client(data_in):
    (name, max_length, experimental) = data_in
    internal_report = []
    my_id = os.getpid()

    def Report(*args):
        args = [ str(x) for x in args ]
        out = ' '.join(args) 
        print my_id, out
        internal_report.append(out)

    Report('pid %d host %s' % (my_id, os.uname()[ 1 ]))
    Report('program %s max_length %d experimental %d' % (
                    name, max_length, experimental))

    try:
        data_out = Do_Program(name, Report, 0, max_length, experimental)
        Report("normal exit")
    except M5_Bug, e:
        Report('m5 bug, move on')
        data_out = { "summary" : "fail %s" % repr(e) }
    except Exception, e:
        out = traceback.format_exc()
        data_out = { "summary" : "fail %s" % repr(e) }
        Report('exception %s' % repr(e))
        Report(out)
   
    data_out[ "report" ] = internal_report

    return data_out
   
def Prepare_Main():
    collect = dict()
    for name in file("%s/list.out" % BENCHMARKS):
        for ml in [2, 4, 8, 16]:
            for experimental in [False]:
                data_in = (name.strip(), ml, experimental)
                collect[ data_in ] = pickleserver.TODO

    pickle.dump(collect, file("tasks.dat", "wb"))

def Client_Main():
    server = "localhost"
    if len(sys.argv) == 2:
        server = sys.argv[ 1 ]
    elif len(sys.argv) != 1:
        print 'Error - give the server hostname as a parameter.'
        return

    def run():
        pickleserver.clientRun(server, Client)
 
    reactor.addSystemEventTrigger('after', 'startup', run)
    reactor.run()

def Make_Debug_Dir(name, max_length, use_temp_dir):
    debug_dir = "%s/debug/%s-%u" % (BENCHMARKS, name, max_length)
    if use_temp_dir:
        debug_dir = '/tmp/ytmp-%d' % os.getpid()

    try:
        os.mkdir(debug_dir)
    except:
        pass
    return debug_dir

def Do_Program(name, Report, debug_level, max_length, experimental):
    Report('-' * 70)
    bin = "%s/bin/%s" % (BENCHMARKS, name)
    use_temp_dir = ( debug_level <= 0 )
    debug_dir = Make_Debug_Dir(name, max_length, use_temp_dir)
    cfg_file = "%s/cfg" % (debug_dir)
    udg_file = "%s/cfg.udg" % (debug_dir)
    trace_file = "%s/trace" % (debug_dir)

    fields = name.split('-')
    if fields[ 0 ] == 'alpha':
        m5_path = "m5/build/ALPHA_SE"
        nm_bin = "alpha-unknown-linux-gnu-nm"
    elif fields[ 0 ] == 'arm':
        m5_path = "m5/build/ARM_SE"
        nm_bin = "arm-9tdmi-linux-gnu-nm"
    else:
        assert False, "Unknown platform '%s'" % fields[ 0 ]

    if fields[ 1 ] in BLACKLIST:
        assert False, "%s is blacklisted" % fields[ 1 ]
    if fields[ 2 ] != '3':
        assert False, "only doing 3's"

    assert not experimental

    return_data = { 
            "longest" : 0,
            "process" : [],
        }
    cmd = "./makecfg %s %s %s %s %s %s" % (bin, cfg_file, 
                udg_file, trace_file, nm_bin, m5_path)
    Report(cmd)
    rc = os.system(cmd)
    if ( rc != 0 ):
        Report("Unable to makecfg for benchmark %s" % name)
        assert False, "fail makecfg"
        
    Report('-' * 70)
    # Read input data
    cfg = readcfg.Read_CFG(cfg_file)
    cap_copies = []
    exe_trace_in = []
    for line in file(trace_file, 'rt'):
        pc = int(line, 16)
        exe_trace_in.append(pc)
    exe_trace_in.pop() # remove last instruction
    loop_header_list = readcfg.Identify_Loops(cfg)
    #return_data[ "cfg" ] = cfg

    # Solve IPET
    model = ilp.ILP_Solver("%s/model.%s.iio" % (debug_dir, name))
    iio_wcet = ipet.Simple_IPET(cfg, model)

    Report('program', name)

    # scan edges
    (vertexes, edges) = readcfg.Find_V_E(cfg)
    edges.sort(key=lambda e: e.pc)
    first_pc = 1 << 31
    nonzero_edges = []
    num_conditional_branches = 0
    use_edge = dict()
    for e in edges:
        e.superblock = None
        e.tried = DO_NOT_TOUCH
        e.is_dirty = True

        if not use_edge.has_key(e.pc):
            if (e.pc != None) and (e.cost > 0):
                first_pc = min(e.pc, first_pc)
                use_edge[ e.pc ] = e
                nonzero_edges.append(e)
                e.superblock = None

            if e.branch_type == BT_Conditional:
                num_conditional_branches += 1

            if e.branch_type == BT_None:
                # Join up with successor
                assert len(e.end.exits) <= 1, ( 
                    "bt == None but there are %d exits @ bb %x?" %
                        (len(e.end.exits), e.pc))
                if len(e.end.exits) == 1:
                    e.superblock_next = e.end.exits[ 0 ]

            elif e.branch_type != BT_Serialising:
                if len(e.end.exits) == 1:
                    # One-way branch - forced into superblock form
                    e.superblock_next = e.end.exits[ 0 ]
                else:
                    # Allow the algorithm to decide on superblock layout
                    e.tried = NOT_TRIED

    # Ready to start
    Report('-' * 70)
    Report('iio wcet', iio_wcet)

    def Run_Capture_Cycle(label):
        # Debugging settings
        cap_copies.append("")
        if (debug_level < 2) and len(cap_copies) > 1:
            cap_copies.pop()
        seq_num = len(cap_copies)
        cap_in_file = "%s/cap.%04d.in" % (debug_dir, seq_num)
        cap_out_file = "%s/cap.%04d.out" % (debug_dir, seq_num)
        cap_dat_file = "%s/cap.%04d.dat" % (debug_dir, seq_num)
        log_base = '%s/m5.%04d' % (debug_dir, seq_num)
        helper_out = "%s/helper.%04d" % (debug_dir, seq_num)
        m_udg_out = "%s/measure.%04d.udg" % (debug_dir, seq_num)
        w_udg_out = "%s/wcet.%04d.udg" % (debug_dir, seq_num)
        cap_copies[ -1 ] = cap_out_file

        # make capture program for current CFG
        special_superblock_set = set()
        cap = file(cap_in_file, "wt")
        local_max_len = 0
        save_count = calc_count = 0
        for start_e in nonzero_edges:
            out = [ "#\n" * 4, "STA %x\n" % start_e.pc ]
            e = start_e
            superblock = Superblock(start_e)

            first_pc = start_e.pc
            last_pc = start_e.end_pc
            is_dirty = start_e.is_dirty
            edge_count = dict()
            size = 0

            while (( e.superblock_next != None )
            and ( edge_count.get(e, 0) < 
                        e.transition_limit )
            and ( size < max_length )):
                e1 = next_e = e.superblock_next
                edge_count[ e ] = edge_count.get(e, 0) + 1

                is_dirty = is_dirty or next_e.is_dirty

                # Just in case the next block is not real..
                next_e = Trace_Real(next_e)
                if next_e.end.is_exit:
                    break

                # Other branches?
                other_paths = []
                ok = False
                for e2 in e.end.exits:
                    if e2 == e1:
                        ok = True
                    else:
                        other_paths.append(Trace_Real(e2))
                assert ok

                if ( e.branch_type == BT_Conditional ):
                    out.append("# conditional branch at %x\n" % e.end_pc)
                    if ( next_e.pc == e.end_pc ):
                        if len(other_paths) == 0:
                            out.append("# has only 1 path\n")
                            out.append("BNT %x\n" % first_pc)
                        else:
                            out.append("BNT %x\n" % other_paths[ 0 ].pc)
                    else:
                        out.append("BTK %x\n" % next_e.pc)
                    superblock.Add_Side_Exit(e, other_paths)
                    size += 1
                elif ( e.branch_type == BT_Unconditional ):
                    out.append("# jump at %x\n" % e.end_pc)
                    out.append("JMP %x\n" % next_e.pc)

                    superblock.Add_Side_Exit(e, other_paths)
                    size += 1
                elif ( e.branch_type == BT_Serialising ):
                    out.append("# serialising instruction at %x\n" % e.end_pc)
                    out.append("JMP %x\n" % first_pc)
                    superblock.Add_Side_Exit(e, other_paths)
                    size += 1
                else:
                    superblock.Add_No_Exit(e)
                    assert next_e.pc == e.end_pc, (
                        "e.pc = %x  next_e.pc = %x   e.end_pc = %x" %
                            (e.pc, next_e.pc, e.end_pc))
                e = next_e
                if ( e.end_pc != None ):
                    last_pc = e.end_pc

            is_dirty = is_dirty or e.is_dirty

            if size == 0:
                out.append("# trivial superblock\n")
                if start_e.branch_type != BT_None:
                    out.append("# fake branch prediction\n")
                    out.append("BNT %x\n" % first_pc)
                last_pc = start_e.end_pc - 4
            else:
                if e.branch_type == BT_None:
                    out.append("# branchless basic block\n")
                    # Just END
                    last_pc -= 4
                else:
                    out.append("# final section\n")
                    # Add a prediction for the final branch
                    out.append("BNT %x\n" % first_pc)
                    last_pc -= 4

            out.append("END %x\n" % last_pc)

            superblock.Add_No_Exit(e)

            if e.branch_type == BT_Serialising:
                out.append("# serialising branch\n")
                special_superblock_set.add(first_pc)
                superblock.Add_Followers(
                        [ Trace_Real(x) for x in e.end.exits 
                            if x.pc != None ])

            if is_dirty:
                cap.write(''.join(out))
                start_e.superblock = superblock
                calc_count += 1
            else:
                save_count += 1

            local_max_len = max(local_max_len, size)

        cap.close()
        return_data[ "longest" ] = max(
            return_data[ "longest" ], local_max_len)

        # Dirty edges are now clean
        for e in edges:
            e.is_dirty = False

        # run M5 to capture data about trivial superblocks
        vt_parameters = {
            'cmd' : bin,
            'meta_in' : cap_in_file,
            'meta_out' : cap_out_file,
            'max_length' : local_max_len + 1,
            'warnings_out' : '%s/warn_log.txt' % debug_dir,
            'start_pc' : first_pc,
            'capture_timings' : True,
            'runtime' : 0,
            }
        flags = "GDBAcc"
        if debug_level > 3:
            flags = "O3CPUAll"
        elif debug_level > 2:
            flags = "Fetch,Exec,IEW,Commit"

        pickle.dump(vt_parameters, file(cap_dat_file, "wb"))
        cmd = ("%s/%s --trace-flags=%s --remote-gdb-port=0 "
                "%s/../../configs/example/vt1.py --caches --detailed -c %s "
            ">%s.stdout.txt 2>%s.stderr.txt" % (m5_path, M5_CMD, 
                    flags, m5_path, cap_dat_file, log_base, log_base))
        #Report(cmd)
        rc = os.system(cmd)
        if rc != 0:
            raise M5_Bug()

        # readback data
        cap = file(cap_out_file, "rt")
        start_e = None
        superblock = None
        is_special = False
        for line in cap:
            fields = line.split()
            key = fields[ 0 ]
            value = int(fields[ -1 ], 16)
            if (( value == 0 ) or value > ( 1 << 63 )):
                value = None

            if key == "STA":
                start_e = use_edge[ value ]
                superblock = start_e.superblock
                is_special = ( start_e.pc in special_superblock_set )

            elif key == "LEN":
                assert False, "Protocol error, LEN present?"
                
            elif key in ("BTK", "BNT", "JMP", "END"):

                if is_special:
                    # Fakery for serialising superblocks - only the END
                    # matters for these.
                    if value != None:
                        superblock.Add_Side_Exit_Cost(value)
                        superblock.Add_Side_Exit_Cost(value)
                else:
                    superblock.Add_Side_Exit_Cost(value)

        cap.close()

        if (debug_level < 1):
            debug_out = None 
        else:
            debug_out = file(helper_out, 'wt')
            debug_out.write("To calculate: %d   Saved: %d\n" % (calc_count, 
                        save_count))

        if (debug_level >= 1):
            mt1 = Check_Superblocks(cfg, exe_trace_in, debug_out,
                    file(m_udg_out, 'wt'))

        model = ilp.ILP_Solver("%s/model.%04d" % (debug_dir, seq_num))
        try:
            wcet = Superblock_IPET(cfg, model, debug_out,
                file(w_udg_out, 'wt'))
        except ilp.ILPTimeoutError:
            Report("%s: ILP timeout caught" % label)
            return 1 << 24

        if (debug_level >= 1):
            Report("%s: sn %04d, measured %d, wcet %d" % (label, 
                        seq_num, mt1, wcet))
            assert wcet >= mt1
        else:
            Report("%s: sn %04d, wcet %d" % (label, seq_num, wcet))
        return wcet

    # form trivial superblocks & compute WCET
    cur_wcet = triv_wcet = Run_Capture_Cycle("triv")
    Report('triv wcet', triv_wcet)
    stop_counter = 0

    # now, do a 1st pass 
    Report("First pass - WCET initially %d" % cur_wcet)
    improvement = ( max_length > 0 )
    while improvement:

        improvement = False
        for best_e1 in edges:
            if best_e1.tried != NOT_TRIED:
                continue
            if best_e1.frequency != 0:
                # Tentative assignment made here
                best_e1.tried = STAGE_ONE
                best_freq = 0
                best_e2 = None
                for e in best_e1.end.exits:
                    if e.frequency >= best_freq:
                        best_e2 = e
                        best_freq = e.frequency
                
                if best_e2 != None:
                    best_e1.superblock_next = best_e2
                    best_e1.transition_limit = 1
                    Dirty(best_e1)
                    improvement = True # It's possible...

        if improvement:
            new_wcet = Run_Capture_Cycle("stage_one")
            if new_wcet < cur_wcet:
                improvement = True
            cur_wcet = new_wcet # Note: WCET may have increased

    Report("Second pass - WCET initially %d" % cur_wcet)

    while ((stop_counter < NON_IDEAL_TOLERANCE) and (max_length > 0)):

        # Update frequencies
        cur_wcet = Run_Capture_Cycle("undo")

        # find non-final edge with highest frequency
        best_freq = 0
        best_e1 = None
        best_e2 = None
        for e in edges:
            if e.tried == DO_NOT_TOUCH:
                continue
            h = e.frequency
            if h > best_freq:
                best_e1 = e
                best_freq = h

        if best_e1 == None:
            Report("Stop - out of options")
            break

        # Read to test new branch settings
        best_e1.tried = DO_NOT_TOUCH # it's final now
        restore_next = best_e1.superblock_next 
        restore_limit = best_e1.transition_limit
        def Restore():
            # Restore returns to old branch settings
            best_e1.superblock_next = restore_next
            best_e1.transition_limit = restore_limit
            Dirty(best_e1)

        # find best successor edge
        best_e2_candidates = []
        for e in best_e1.end.exits:
            best_e2_candidates.append(e)

        best_e2_candidates.sort(key=lambda e: -e.frequency)
        if len(best_e2_candidates) > MAX_CANDIDATES:
            Report("Many candidates here! %d" % len(best_e2_candidates))
            best_e2_candidates = best_e2_candidates[ : MAX_CANDIDATES ]

        if len(best_e2_candidates) > 1:
            total_frequency = 0
            for e in best_e2_candidates:
                total_frequency += e.max_frequency

            total_frequency = float(total_frequency) / len(best_e2_candidates)
            bias = 1.0
            for e in best_e2_candidates:
                bias = min(bias,
                    abs((2.0 * (float(e.max_frequency) / total_frequency)) - 1.0))
            Report("Frequency %d, bias score: %1.3f" % (total_frequency, bias))
            # bias score 0.0 - branches are unbiased (all equally likely)
            # bias score 1.0 - branches are strongly biased
            

        scan = []
        for best_e2 in best_e2_candidates:
            best_e1.superblock_next = best_e2
            e = Trace_Real(best_e2)
            if e.end.is_exit:
                # Starting edge best_e1 is not suitable because it
                # leads directly to the exit.
                continue

            Report('Trying join of %x to %x (E%06x -> E%06x [E%06x])' % (
                    best_e1.pc, e.pc, best_e1.id, best_e2.id, e.id))
           
            todo = range(1, max_length)
            todo.reverse()
            while len(todo) != 0:
                i = todo.pop()
                # re-evaluate WCET
                best_e1.transition_limit = i
                Dirty(best_e1)

                scan.append((Run_Capture_Cycle(str(i)), i, best_e2))

                if ((len(todo) != 0)
                and (i >= STOP_HEURISTIC_POINT) and STOP_HEURISTIC 
                and (scan[ -1 ][ 0 ] == scan[ -2 ][ 0 ])):
                    # Heuristic - save time if a fixed point has been reached
                    # by skipping to the end
                    todo = [ max_length - 1 ]

            #Report('scan:', [ x[ 0 ] for x in scan ])

        Restore()

        if len(scan) == 0:
            Report("No suitable candidates here")
            continue

        def Scan_Sort(a, b):
            if a[ 0 ] < b[ 0 ]:
                return -1
            if a[ 0 ] > b[ 0 ]:
                return 1
            if a[ 1 ] < b[ 1 ]:
                return 1
            if a[ 1 ] > b[ 1 ]:
                return -1
            return 0

        scan.sort(cmp=Scan_Sort)
        (new_wcet, i, best_e2) = scan[ 0 ]
        best_e1.superblock_next = best_e2
        best_e1.transition_limit = i
        if best_e2 != best_e2_candidates[ 0 ]:
            Report("Chose alternate branch")

        if new_wcet > cur_wcet:
            # Incorrect move made, just undo for now
            Restore()
            improve_factor = 1.0
            Report('WCET worse by at least %d, undo - %d' % (
                    new_wcet - cur_wcet, cur_wcet))
        else:
            improve_factor = float(new_wcet) / float(cur_wcet)
            Report('WCET better by %d, picking %d, now %d, factor %1.3f' % (
                    cur_wcet - new_wcet, i, new_wcet, improve_factor))
            cur_wcet = new_wcet

        if ( improve_factor <= MINIMUM_CHANGE_FACTOR ):
            stop_counter = 0
        else:
            stop_counter += 1
        return_data[ "process" ].append(cur_wcet)

    if stop_counter >= NON_IDEAL_TOLERANCE:
        Report("Stop - not making enough improvement")
        return_data[ "unimprovement" ] = stop_counter

    Report('ncb', num_conditional_branches)
    Report('iio wcet', iio_wcet)
    Report('triv wcet', triv_wcet)
    Report('o3+vtc wcet', cur_wcet)
    r = (100.0 * (( float(iio_wcet) / float(cur_wcet) ) - 1.0 ))
    Report('reduction %1.1f%%' % r)
    return_data.update({
        "iio" : iio_wcet,
        "ncb" : num_conditional_branches,
        "triv" : triv_wcet,
        "o3vtc": cur_wcet,
        "summary" : r,
        })
    if use_temp_dir:
        try:
            shutil.rmtree(debug_dir)
        except:
            pass
    return return_data

def Trace_Real(e):
    start = e
    while (e.pc == None) and not e.end.is_exit:
        assert e.branch_type == BT_None
        assert len(e.end.exits) <= 1
        e = e.superblock_next
        assert e != None, ( "Unable to find real edge from start E%06x" %
                            (start.id))
    return e

def Dirty(e):
    e.is_dirty = True
    e = Trace_Real(e)
    if e != None:
        e.is_dirty = True


if ( __name__ == "__main__" ):
    sys.exit(Main())
   
