#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <assert.h>

#define M5_CMD              "m5.opt"
#define MAX_TICK            1e9     // one Billion cycles

#include "makecfg.hh"
#include "unix.hh"


static void Output_Dump ( Make_CFG * m, Symbol_Table * symtab,
            std::string debug_dir, std::string num )
{
    std::string     output;
    std::string     udg_output;
    FILE *          out_fd ;

    output = debug_dir + "/cfg" + num;
    out_fd = fopen ( output . c_str () , "wt" ) ;
    if ( out_fd == 0L )
    {
        perror ( "Unable to open output file." ) ;
        return ;
    }
    m -> Export ( out_fd ) ;
    fclose ( out_fd ) ;

    udg_output = debug_dir + "/cfg" + num + ".udg";
    out_fd = fopen ( udg_output . c_str () , "wt" ) ;
    if ( out_fd == 0L )
    {
        perror ( "Unable to open UDG output file." ) ;
        return ;
    }
    m -> Export_UDG ( out_fd , symtab ) ;
    fclose ( out_fd ) ;
}

bool Do_Make_CFG_and_Def_Use ( 
                std::string program1 ,
                std::string debug_dir ,
                std::string nm_bin ,
                std::string m5_path )
{
    char            cmd [ BUFSIZ ] ;
    char            fifo_name[ 3 ][ BUFSIZ ] ;
    int             uid = getpid () ;
    Make_CFG *      m = new Make_CFG () ;
    Symbol_Table *  symtab = 0L ;
    FILE *          m5fd[ 3 ] = { 0L, 0L, 0L };
    FILE *          fifo_fd[ 3 ] = { 0L, 0L, 0L };
    FILE *          out_fd = 0L ;
    ulongint        main_end , main_start , main_size ;
    bool            rc = false ;
    bool            fail = false ;
    bool            src ;
    ulongint        i;
    std::string     script1 = debug_dir + "/todo1.dat";
    std::string     script2 = debug_dir + "/todo2.dat";
    std::string     script3 = debug_dir + "/todo3.dat";

    cmd [ BUFSIZ - 1 ] = '\0' ;
    for (i = 0; i < 3; i++) {
        fifo_name[ i ][ BUFSIZ - 1 ] = '\0' ;
        snprintf ( fifo_name[ i ] , BUFSIZ - 1 , "/tmp/%d-%u.fifo" , 
                    uid, (unsigned) i ) ;
    }

    do {
        fprintf ( stderr , "Reading symbols...\n" );
        symtab = Do_Read_Symbols ( program1 , nm_bin ) ;
        if ( symtab == 0L )
        {
            break ;
        }

        src = symtab -> Get_Record ( "trigger_level" , 
                                main_start , main_size ) ;
        main_end = main_start + main_size ;

        if ( ! src )
        {
            fprintf ( stderr , "Error reading symbols "
                "(trigger_level not found)\n" ) ;
            break ;
        }
       
        fprintf ( stderr , "trigger_level at 0x%"FX" to 0x%"FX"...\n",
                main_start, main_end );
        fprintf ( stderr , "Creating simulators and FIFOs...\n" );

        for (i = 0; i < 3; i ++) {
            if ( mkfifo ( fifo_name[ i ] , 0600 ) != 0 ) {
                perror ( "Error creating FIFO" ) ;
                fail = true;
                break;
            }
            const char * script_str = NULL;
            switch (i) {
                case 0:     script_str = script1.c_str(); break;
                case 1:     script_str = script2.c_str(); break;
                case 2:     script_str = script3.c_str(); break;
                default:    assert(0); break;
            }

            snprintf ( cmd , BUFSIZ - 1 ,
                    "%s/" M5_CMD " --trace-flags=Exec,ExecFTR "
                    "--remote-gdb-port=0 "
                    "--trace-file %s bench.py --maxtick %"FU" "
                    "%s" , m5_path . c_str () , 
                        fifo_name[ i ] , 
                        (ulongint) MAX_TICK,
                        script_str ) ;
            fprintf ( stderr , "M5: %s\n" , cmd ) ;

            m5fd[ i ] = popen ( cmd , "r" ) ;
            if ( m5fd[ i ] == 0L ) {
                perror ( "Unable to start M5." ) ;
                fail = true;
                break;
            }

            fifo_fd[ i ] = fopen ( fifo_name[ i ] , "rb" ) ;
            if ( fifo_fd[ i ] == 0L ) {
                perror ( "Unable to open FIFO." ) ;
                fail = true;
                break;
            }
        }

        if (fail) break;

        fprintf ( stderr , "\nReading trace...\n" );

        // Initial CFG capture 
        m -> Read_Trace ( fifo_fd[ 2 ] , 
                    main_start , main_end , 0L ) ;
        fclose ( m5fd[ 2 ] ) ;
        fclose ( fifo_fd[ 2 ] ) ;
        m5fd[ 2 ] = fifo_fd[ 2 ] = 0L ;
        //wait ( 0L ) ;

        fprintf ( stderr , "Making CFG\n" );
        m -> Make () ;

        Output_Dump ( m, symtab, debug_dir, "1" );


        fprintf ( stderr , "Begin def-use analysis...\n" );

        // Def-use capture
        if (!m->Hacked_Up_Def_Use_Analysis (
                fifo_fd[ 0 ], fifo_fd[ 1 ], main_start, main_end)) {
            break;
        }

        fprintf ( stderr , "Output stage...\n" );
        for (i = 0; i < 2; i ++) {
            fclose ( m5fd[ i ] ) ;
            fclose ( fifo_fd[ i ] ) ;
            m5fd[ i ] = fifo_fd[ i ] = 0L ;
            //wait ( 0L ) ;
        }

        Output_Dump ( m, symtab, debug_dir , "2" );

        fprintf ( stderr , "Done (success)\n" );
        rc = true ;
    } while ( 0 ) ;

    delete m ;
    if ( symtab != 0L ) delete symtab ;
    if ( out_fd != 0L ) fclose ( out_fd ) ;

    for (i = 0; i < 3; i ++) {
        if ( m5fd[ i ] != 0L ) fclose ( m5fd[ i ] ) ;
        if ( fifo_fd[ i ] != 0L ) fclose ( fifo_fd[ i ] ) ;
        unlink ( fifo_name[ i ] ) ;
    }

    return rc ; 
}

void __assert_fail (__const char *__assertion, __const char *__file,
               unsigned int __line, __const char *__function)
{
    fprintf(stderr, "\nBang! Assertion failed:\n"
            "  %s\n"
            "  %s:%u in %s\n"
            "Attach your debugger to pid %d now!\n", 
            __assertion, __file, __line, __function,
            getpid());
    while (1) {
        sleep ( 1 );
        sleep ( 1 );
    }
}


int main ( int argc , char * argv [] )
{
    if ( argc != 5 )
    {
        fprintf ( stderr , "Bad parameter area.\n" );
        return 1 ;
    }
    if ( Do_Make_CFG_and_Def_Use ( argv [ 1 ] , 
            argv [ 2 ] , argv [ 3 ] , argv [ 4 ] ) )
    {
        return 0 ;
    }
    fprintf ( stderr , "%s: failed.\n" , argv [ 0 ] ) ;
    return 1 ;
}

